
package org.mule.yammer.config;

import java.io.OutputStreamWriter;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.Date;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.log4j.Logger;
import org.mule.api.MessagingException;
import org.mule.api.MuleContext;
import org.mule.api.MuleEvent;
import org.mule.api.MuleException;
import org.mule.api.adapter.OAuth2Adapter;
import org.mule.api.callback.HttpCallback;
import org.mule.api.context.MuleContextAware;
import org.mule.api.lifecycle.Initialisable;
import org.mule.api.lifecycle.Startable;
import org.mule.api.lifecycle.Stoppable;
import org.mule.api.processor.MessageProcessor;
import org.mule.config.i18n.MessageFactory;
import org.mule.util.IOUtils;
import org.mule.yammer.YammerConnector;
import org.mule.yammer.config.spring.DefaultHttpCallback;


/**
 * A {@code YammerConnectorOAuth2Adapter} is a wrapper around {@link YammerConnector } that adds OAuth capabilites to the pojo.
 * 
 */
public class YammerConnectorOAuth2Adapter
    extends YammerConnectorHttpCallbackAdapter
    implements OAuth2Adapter, MuleContextAware, Initialisable, Startable, Stoppable
{

    private final static Pattern AUTH_CODE_PATTERN = Pattern.compile("code=([^&]+)");
    private final static Pattern ACCESS_CODE_PATTERN = Pattern.compile("\"token\":\"([^&]+?)\"");
    private final static Pattern EXPIRATION_TIME_PATTERN = Pattern.compile("\"expires_in\":([^&]+?),");
    private MuleContext muleContext;
    private String oauthVerifier;
    private String redirectUrl;
    private HttpCallback oauthCallback;
    private String accessToken;
    private Date expiration;
    private final static Logger LOGGER = Logger.getLogger(YammerConnectorOAuth2Adapter.class);

    /**
     * Sets muleContext
     * 
     * @param value Value to set
     */
    public void setMuleContext(MuleContext value) {
        this.muleContext = value;
    }

    /**
     * Retrieves oauthVerifier
     * 
     */
    public String getOauthVerifier() {
        return this.oauthVerifier;
    }

    /**
     * Sets oauthVerifier
     * 
     * @param value Value to set
     */
    public void setOauthVerifier(String value) {
        this.oauthVerifier = value;
    }

    /**
     * Retrieves redirectUrl
     * 
     */
    public String getRedirectUrl() {
        return this.redirectUrl;
    }

    /**
     * Retrieves accessToken
     * 
     */
    public String getAccessToken() {
        return this.accessToken;
    }

    /**
     * Sets accessToken
     * 
     * @param value Value to set
     */
    public void setAccessToken(String value) {
        this.accessToken = value;
    }

    /**
     * Sets expiration
     * 
     * @param value Value to set
     */
    public void setExpiration(Date value) {
        this.expiration = value;
    }

    public void start()
        throws MuleException
    {
        super.start();
        oauthCallback.start();
        redirectUrl = oauthCallback.getUrl();
    }

    public void stop()
        throws MuleException
    {
        super.stop();
        oauthCallback.stop();
    }

    public void initialise() {
        super.initialise();
        oauthCallback = new DefaultHttpCallback(new YammerConnectorOAuth2Adapter.OnOAuthCallbackMessageProcessor(), muleContext, getDomain(), getLocalPort(), getRemotePort(), "yammerCodeRetrievingPath", getAsync());
    }

    public String getAuthorizationUrl() {
        StringBuilder urlBuilder = new StringBuilder();
        urlBuilder.append("https://www.yammer.com/dialog/oauth");
        urlBuilder.append("?");
        urlBuilder.append("response_type=code&");
        urlBuilder.append("client_id=");
        urlBuilder.append(getConsumerKey());
        urlBuilder.append("&redirect_uri=");
        urlBuilder.append(redirectUrl);
        LOGGER.info(("OAUth 2 authorization url: " + urlBuilder));
        return urlBuilder.toString();
    }

    public void fetchAccessToken() {
        try {
            HttpURLConnection conn = ((HttpURLConnection) new URL("https://www.yammer.com/oauth2/access_token.json").openConnection());
            conn.setRequestMethod("POST");
            conn.setDoOutput(true);
            StringBuilder builder = new StringBuilder();
            builder.append("code=");
            builder.append(URLEncoder.encode(oauthVerifier, "UTF-8"));
            builder.append("&client_id=");
            builder.append(URLEncoder.encode(getConsumerKey(), "UTF-8"));
            builder.append("&client_secret=");
            builder.append(URLEncoder.encode(getConsumerSecret(), "UTF-8"));
            builder.append("&grant_type=");
            builder.append(URLEncoder.encode("authorization_code", "UTF-8"));
            builder.append("&redirect_uri=");
            builder.append(URLEncoder.encode(redirectUrl, "UTF-8"));
            OutputStreamWriter out = new OutputStreamWriter(conn.getOutputStream());
            out.write(builder.toString());
            out.close();
            String response = IOUtils.toString(conn.getInputStream());
            Matcher matcher = ACCESS_CODE_PATTERN.matcher(response);
            if (matcher.find()&&(matcher.groupCount()>= 1)) {
                accessToken = URLDecoder.decode(matcher.group(1), "UTF-8");
            } else {
                throw new Exception(String.format("OAuth access token could not be extracted from: %s", response));
            }
            Matcher expirationMatcher = EXPIRATION_TIME_PATTERN.matcher(response);
            if (expirationMatcher.find()&&(expirationMatcher.groupCount()>= 1)) {
                Long expirationSecsAhead = Long.parseLong(expirationMatcher.group(1));
                expiration = new Date((System.currentTimeMillis()+(expirationSecsAhead* 1000)));
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public boolean hasTokenExpired() {
        return ((expiration!= null)&&expiration.before(new Date()));
    }

    public void reset() {
        expiration = null;
        oauthVerifier = null;
        accessToken = null;
    }

    private class OnOAuthCallbackMessageProcessor
        implements MessageProcessor
    {


        public MuleEvent process(MuleEvent event)
            throws MuleException
        {
            try {
                oauthVerifier = extractAuthorizationCode(event.getMessageAsString());
            } catch (Exception e) {
                throw new MessagingException(MessageFactory.createStaticMessage("Could not extract OAuth verifier"), event, e);
            }
            return event;
        }

        private String extractAuthorizationCode(String response)
            throws Exception
        {
            Matcher matcher = AUTH_CODE_PATTERN.matcher(response);
            if (matcher.find()&&(matcher.groupCount()>= 1)) {
                return URLDecoder.decode(matcher.group(1), "UTF-8");
            } else {
                throw new Exception(String.format("OAuth authorization code could not be extracted from: %s", response));
            }
        }

    }

}
